/* 
 MEX function ss=mhgis(MAX,alpha,p,q,n,x,K).
 Computes the truncated hypergeometric function pFq ^alpha(p;q;x*I_n)
 The sum is only over partitions |kappa|<=MAX; such that 
 kappa_1 <= K.
 The only difference between mhgi and mhgis is that mhgis allows for the
 additional parameter K.
 p and q are arrays, so hgi(30,a,[3 4],[5 6 7],3,0.5,9) is 
 2F3^a([3 4];[5 6 7];0.5*I_3) summed over all partitions |kappa|<=30
 such that kappa_1 <=K
 x may be a vector; if so, then a vector is returned.
 Uses the formula 
 J_lambda(x I_n)=x^|lambda| \prod_{(i,j)\in \lambda (n-(i-1)+alpha(j-1))

 by Plamen Koev, February 2004, May 2004.

 Removed recursion, April 2005.

 Added parameter K, August 2005

 Copyright (c) Plamen Koev 2004-2005
*/

#include <mex.h>
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>

void mexFunction( int nlhs, mxArray *plhs[], 
                  int nrhs, const mxArray*prhs[]) {
  int i, MAX, n, nx, np, nq, *l, *lc, j, sl, K;
  double alpha, *x,*p, *q, *s;
  double *z, zn,c,d,e,f,g;

  if (nrhs!=7)
    mexErrMsgTxt("Must have exactly seven input arguments.");
  if (nlhs>1)
    mexErrMsgTxt("Too many output arguments.");
  
  if (!mxIsDouble(prhs[0]) || !mxIsDouble(prhs[1]) || !mxIsDouble(prhs[2]) || !mxIsDouble(prhs[3]) || !mxIsDouble(prhs[4]) || !mxIsDouble(prhs[5]))
    mexErrMsgTxt("All inputs must be double arrays.");

  if (mxGetM(prhs[0])!=1 || mxGetN(prhs[0])!=1)
    mexErrMsgTxt("First input must be a scalar = max size of partition.");
  if (mxGetM(prhs[1])!=1 || mxGetN(prhs[1])!=1)
    mexErrMsgTxt("Second input must be a scalar = alpha.");
  if (mxGetM(prhs[2])>1 && mxGetN(prhs[2])>1)
    mexErrMsgTxt("Third input must be a vector = (a_1,...,a_p).");
  if (mxGetM(prhs[3])>1 && mxGetN(prhs[3])>1)
    mexErrMsgTxt("Forth input must be a vector = (b_1,...,b_q).");
  if (mxGetM(prhs[4])!=1 || mxGetN(prhs[4])!=1)
    mexErrMsgTxt("Fifth input must be a scalar = number of variables.");
  if (mxGetM(prhs[5])!=1 && mxGetN(prhs[5])!=1)
    mexErrMsgTxt("Sixth input must be a vector = x.");
  if (mxGetM(prhs[6])!=1 || mxGetN(prhs[6])!=1)
    mexErrMsgTxt("Seventh input must be a scalar = number of variables.");

  plhs[0]=mxDuplicateArray(prhs[5]);   /* output is the same size as input */
  s=mxGetPr(plhs[0]);

  x=mxGetPr(prhs[5]);
  nx=mxGetNumberOfElements(prhs[5]);  /* size of x */
  p=mxGetPr(prhs[2]);
  np=mxGetNumberOfElements(prhs[2]);  /* size of p */
  q=mxGetPr(prhs[3]);
  nq=mxGetNumberOfElements(prhs[3]);  /* size of q */


  MAX=(int) *mxGetPr(prhs[0]);
  alpha=*mxGetPr(prhs[1]);
  n = (int) *mxGetPr(prhs[4]);
  K= (int) *mxGetPr(prhs[6]);

  for (i=0;i<nx;i++) s[i]=1;

  if (K==0) return;

  l =(int*)mxMalloc(sizeof(int)*(MAX+2));
  memset(l,0,sizeof(int)*(MAX+2));
  lc =(int*)mxMalloc(sizeof(int)*(MAX+2));
  memset(lc,0,sizeof(int)*(MAX+2));
  z=(double*)mxMalloc(sizeof(double)*(MAX+1)*nx);
  memcpy(z+nx,s,sizeof(double)*nx);    /* set to 1 also */

  l[0]=MAX+1; /* purely for convenience */
  sl=0;
  i=1;

  while (i>0) {
     if ((sl<MAX) && (i<=n) && (l[i]<l[i-1]) && (l[i]<K)  ) {
       l[i]++;
       lc[l[i]]++;        /* update conjugate partition */
       sl++;

       zn=1;
       c=-(i-1)/alpha+l[i]-1;
       for (j=0;j<np;j++)  zn*=p[j]+c;
       for (j=0;j<nq;j++)  zn/=q[j]+c;

       d=l[i]*alpha-i;             /* update j_lambda  */
       for (j=1;j<l[i];j++) {
          e=d-j*alpha+lc[j];
          g=e+1;
          zn*=(g-alpha)*e/(g*(e+alpha));
       }
       for (j=1;j<i;j++) {
          f=l[j]*alpha-j-d;
          g=f+alpha;
          e=f*g;
          zn*=(e-f)/(g+e);
       }
       zn*=(n+alpha*c);
       for (j=0;j<nx;j++) {
          z[i*nx+j]*=x[j] * zn;
          s[j]+=z[i*nx+j];
	    }

	    if ((sl<MAX)&&(i<n))
		   memcpy(z+(i+1)*nx,z+i*nx,nx * sizeof(double));
       i++;
     }
     else {
       sl-=l[i];
       for (j=1;j<l[i]+1;j++) lc[j]--;
       l[i]=0;
       i--;
     }
    } /* of while */
 
  mxFree(z);
  mxFree(lc);
  mxFree(l);
}

